/*
             LUFA Library
     Copyright (C) Dean Camera, 2010.
              
  dean [at] fourwalledcubicle [dot] com
      www.fourwalledcubicle.com
*/

/*
  Copyright 2009 Dean Camera (dean [at] fourwalledcubicle [dot] com)
    Additional LoopBack demo code by Opendous Inc. 2010-03-03
    For more info visit: www.Micropendous.org/LoopBack

  This firmware enumerates as a vendor-class device meaning
  the developer must decide and code a communication scheme.
  Examples of arbitrary endpoint as well as control endpoint
  communication are given.
  Search for TODO statements for implementation hints.
  Avoid busy loops as the USB task is not preemptive and USB
  has timing contraints.  Use concurrent programming techniques.
  For example, have one function set a flag when an event occurs.
  Have another function do significant processing only if the flag is set.
  Have both functions run from main()'s for(;;) loop, which should be the
  only indefinite loop in your firmware.

  Permission to use, copy, modify, and distribute this software
  and its documentation for any purpose and without fee is hereby
  granted, provided that the above copyright notice appear in all
  copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaim all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/** \file
 *
 *  Main source file for the LoopBack demo. This file contains the main tasks of the demo and
 *  is responsible for the initial application hardware configuration.
 */
 
#include "LoopBack.h"
#include "circbuf.h"
#define OVERRUN  0x40
#define ROLLOVER 0x80
/* Global Variables */
/* EP_SIZE is defined in Descriptors.h and should be
 * less than or equal to 64 bytes in order to work with Control Endpoint data.
 */

volatile uint8_t dataToSend[EP_SIZE];
volatile uint8_t dataReceived[EP_SIZE];
volatile uint8_t fullFlag; 	// flag indicating that circbuf write overflow
volatile uint8_t writing; 	// flag indicating writing stagingBuffer
volatile uint8_t stagingBuffer[USB_BUFSIZE];
volatile uint16_t outputCount;
volatile uint8_t uartDelay;

/** Main program entry point. This routine configures the hardware required by the application, then
 *  starts the scheduler to run the USB management task.
 */
int main(void)
{
	int i;
	outputCount = 0;
	fullFlag = 0;
	writing  = 0;
	uartDelay = 0;
	for(i = 0; i< USB_BUFSIZE; i++)
		stagingBuffer[i] = i;
		
	CircBufferInit();
	SetupHardware();

	for (;;)
	{
		Main_Task();
		USB_USBTask();
	}
}

/** Configures the board hardware and chip peripherals for the demo's functionality. */
void SetupHardware(void)
{
	/* Disable watchdog if enabled by bootloader/fuses */
	MCUSR &= ~(1 << WDRF);
	wdt_disable();

	/* Disable clock division */
	clock_prescale_set(clock_div_1);

	/* disable JTAG to allow corresponding pins to be used - PF4, PF5, PF6, PF7 */
	/* TODO - remove this if you want to use your JTAG debugger to debug this firmware */
	#if ((defined(__AVR_AT90USB1287__) || defined(__AVR_AT90USB647__) ||  \
			defined(__AVR_AT90USB1286__) || defined(__AVR_AT90USB646__) ||  \
			defined(__AVR_ATmega16U4__)  || defined(__AVR_ATmega32U4__) ||  \
			defined(__AVR_ATmega32U6__)))
		// note the JTD bit must be written twice within 4 clock cycles to disable JTAG
		// you must also set the IVSEL bit at the same time, which requires IVCE to be set first
		// port pull-up resistors are enabled - PUD(Pull Up Disable) = 0
		MCUCR = (1 << JTD) | (1 << IVCE) | (0 << PUD);
		MCUCR = (1 << JTD) | (0 << IVSEL) | (0 << IVCE) | (0 << PUD);
	#endif

	/* Hardware Initialization */
	/* enable Ports based on which IC is being used */
	/* For more information look over the corresponding AVR's datasheet in the
		'I/O Ports' Chapter under subheading 'Ports as General Digital I/O' */
	#if (defined(__AVR_AT90USB162__)  || defined(__AVR_AT90USB82__))
		DDRD = 0;
		PORTD = 0;
		DDRB = 0;
		PORTB = 0;
		DDRC = 0;
		PORTC |= (0 << PC2) | (0 << PC4) | (0 << PC5) | (0 << PC6) | (0 << PC7); //only PC2,4,5,6,7 are pins
		// be careful using PortC as PC0 is used for the Crystal and PC1 is RESET
	#endif

	#if (defined(__AVR_ATmega16U4__)  || defined(__AVR_ATmega32U4__))
		DDRD = 0;
		PORTD = 0;
		DDRB = 0;
		PORTB = 0;
		DDRC = 0;
		PORTC = (0 << PC6) | (0 << PC7); //only PC6,7 are pins
		DDRE = 0;
		PORTE = (0 << PE2) | (0 << PE6); //only PE2,6 are pins
		DDRF = 0;
		PORTF = (0 << PF0) | (0 << PF1) | (0 << PF4) | (0 << PF5) | (0 << PF6) | (0 << PF7); // only PF0,1,4,5,6,7 are pins
	#endif

	#if (defined(__AVR_AT90USB1287__) || defined(__AVR_AT90USB647__) ||  \
			defined(__AVR_AT90USB1286__) || defined(__AVR_AT90USB646__) ||  \
			defined(__AVR_ATmega32U6__))
		DDRA = 0;
		PORTA = 0;
		DDRB 	= 0x0; 		// port b is all inputs
		PORTB	= 0xff; 	// enable pullups PortB 
		DDRC = 0;
		PORTC = 0;
		DDRD = 0;
		PORTD = 0;
		DDRE = ((1 << PE4) | (1 << PE6));	// set PE4, PE6 to HIGH to disable external SRAM, if connected
		PORTE = ((1 << PE4) | (1 << PE6));	// set PE4, PE6 to HIGH to disable external SRAM, if connected
		DDRF = 0;
		PORTF = 0;
	#endif
    
	USB_Init();

	PCMSK0	= 0x3f; 	// enable pin change on all low 6 lines
	PCIFR	= 1;		// clear flag
	PCICR	= (1<<PCIE0); // enable pinchange interrupt
	/* Timer0 is free running at /8 prescale. Every event interrupt captures
	 * the time of the event and puts it in the output buffer.
	 * Output format: <PORTB Pins> <TimerLo>
	 * The overflow interrupt is enabled and the msbit of the port byte.
	*/

    TCCR0A = 0x0;		// free running 8 bit timer
    TCCR0B = 0x03;		// 1 = no prescale, 2 = div by 8; 3 = div by 64 prescale
    TIMSK0 = (1<<TOIE0);// enable overflow interrupt
 // UART Setup, no interrupts   
    UBRR1H  = 0;
    UBRR1L  = 103;	// 9600- 103, 19200 51 , 57600 - 26 , 115.2k - 8, 250k - 3,
	UCSR1B 	= 8;	// TX enable
	UCSR1C	= 0x06;  	// 8 data bits, no parity, 1 stop
	
}

/** Event handler for the USB_Connect event. This indicates that the device is enumerating via the status LEDs and
 *  starts the library USB task to begin the enumeration and USB management process.
 */
void EVENT_USB_Device_Connect(void)
{
	/* Indicate USB enumerating */
//	LEDs_SetAllLEDs(LEDMASK_USB_ENUMERATING);
}

/** Event handler for the USB_Disconnect event. This indicates that the device is no longer connected to a host via
 *  the status LEDs.
 */
void EVENT_USB_Device_Disconnect(void)
{
	/* Indicate USB not ready */
//	LEDs_SetAllLEDs(LEDMASK_USB_NOTREADY);
}

/** Event handler for the USB_ConfigurationChanged event. This is fired when the host sets the current configuration
 *  of the USB device after enumeration, and configures the device endpoints.
 */
void EVENT_USB_Device_ConfigurationChanged(void)
{	

	/* Setup SendDataToHost Endpoint */
	if (!(Endpoint_ConfigureEndpoint(IN_EP, EP_TYPE_BULK,
														ENDPOINT_DIR_IN, EP_SIZE,
														ENDPOINT_BANK_SINGLE)))
	{
	
	}
	
	/* Setup ReceiveDataFromHost Endpoint */
	if (!(Endpoint_ConfigureEndpoint(OUT_EP, EP_TYPE_BULK,
														ENDPOINT_DIR_OUT, EP_SIZE,
														ENDPOINT_BANK_SINGLE)))
	{

	}
}

/** Event handler for the USB_UnhandledControlRequest event. This is used to catch standard, class,
 *  and vendor specific control requests that are not handled internally by the USB library so that
 *  they can be handled appropriately for the application.
 */
void EVENT_USB_Device_UnhandledControlRequest(void)
{
/* Process specific control requests */
	switch (USB_ControlRequest.bRequest)
	{
		default: break;
	}
}


/** Sends data to the host via regular endpoints */
void SendDataToHost(void)
{
	/* Select the IN Endpoint */
	Endpoint_SelectEndpoint(IN_EP);
	
	if (Endpoint_IsINReady())
	{
		/* Check if IN Endpoint is ready for Read/Write */
		while (Endpoint_IsReadWriteAllowed())
		{
			/* Write data to the host */
			Endpoint_Write_Byte(stagingBuffer[outputCount++]);
			if(outputCount == USB_BUFSIZE)
			{
				writing = 0;
				outputCount = 0;
				break;
			}
		}
		Endpoint_ClearIN();
	}

}

/** Reads data from the host via regular endpoints */
void ReceiveDataFromHost(void)
{
	uint8_t temp; // ??? do something with the data?
	
	/* Select the OUT stream endpoint */
	Endpoint_SelectEndpoint(OUT_EP);

	/* Check if the current endpoint can be read */
	if (Endpoint_IsOUTReceived())
	{
		
		/* Write the sample to the buffer */
		while(Endpoint_IsReadWriteAllowed())
		{
			// TODO - instead of a temporary byte, read data into your data structure
			temp=Endpoint_Read_Byte();
		}
		Endpoint_ClearOUT();
	}

}

/** Main_Task will only Send if USB is connected */
void Main_Task(void)
{
SAMPLE sample;

	if ((UCSR1A & 0x20) == 0x20) // txd reg empty
		uartDelay++;
		if((uartDelay & 0x80) == 0x80)
				UDR1 = 0xff;
	if (USB_DeviceState != DEVICE_STATE_Configured) 
		return;
		
	if(!writing) // wait for write to finish ...
	{
		while(!CircBufferDeque(&sample))
		{
			if(fullFlag)
			{
				sample.value |= OVERRUN;
				fullFlag = 0;
			}
// stagingBuffer is used to insure that both bytes can be written
			stagingBuffer[outputCount++] = sample.value;
			stagingBuffer[outputCount++] = sample.timerL;

			if(outputCount == USB_BUFSIZE )
			{
				writing  = 1;
				outputCount = 0;
				SendDataToHost();
				break;
			}
		}
	}else
		SendDataToHost();
}

ISR(TIMER0_OVF_vect)
{ 
	uint8_t sreg;
	SAMPLE sample;

	sreg 			= SREG;
	cli();
	sample.timerL	= TCNT0; //& 0xff;
	sample.value 	= (PINB & 0x3f) + ROLLOVER; // set high bit to indicate a counter rollover
	fullFlag += CircBufferEnque(&sample);

	SREG = sreg;
}
ISR(PCINT0_vect)
{
	uint8_t sreg;
	SAMPLE sample;
	
	sreg 			= SREG;
	cli();
	sample.timerL	= TCNT0;
	sample.value 	= PINB & 0x3f;
	fullFlag += CircBufferEnque(&sample);

	SREG = sreg;
}


